#pragma once

#include <iostream>
#include <stdexcept>
#include <boost/shared_ptr.hpp>

// MathData
#include <NemoMath/Vector.h>

namespace cca {

/**
 * Base for data transfer object, representing domain-specific data.
 * @defgroup DTO DataTransferObjects
 */
class DataTransferObject;
typedef boost::shared_ptr<DataTransferObject> DataTransferObjectPtr;
typedef boost::shared_ptr<void> VoidPtr;

/**
 * Base data transfer object, representing domain-specific data. All inheriting
 * classes should use the International System of Units (SI) as default data
 * representation.
 * @ingroup DTO
 * @todo importers, exporters
 * @todo serialze, deserialize
 * @todo toXML, fromXML
 * @todo toCSV, fromCSV
 * @todo overload operators (+=, -=, *= (), [])
 * @todo provide friend operators (+, -, *, /, !=, ==)
 */
class DataTransferObject {

public:

    /**
     * Prints the DataTransferObject
     * @Return DataTranserObject as String
     */
    virtual std::string print() const;

    /**
     * Stream operator
     * @param os Output sream to attach to
     * @param dto The DataTransferObject to print
     * @return Modified output stream
     */
    friend std::ostream
    & operator<<(std::ostream& os, const DataTransferObject& dto);

protected:
    /**
     * Constructor
     * Initializes DTO with specified dimension.
     */
    DataTransferObject();

    virtual ~DataTransferObject();
};

/**
 * Base for data transfer object, representing domain-specific data.
 * @defgroup DTO DataTransferObjects
 */
class ContainingDoubles {

public:
    /**
     * Get dimension / number of values
     * @return Number of contained values
     */
    virtual unsigned int getDimension() const;

    /**
     * Returns single value as double. Only possible, if dimension is 1.
     * @return Single contained value
     * @todo Throw exceptions if dimension!-1
     */
    virtual double asDouble() const;

    /**
     * Returns single value as double
     * @param index Index of the joint value
     * @return Actual value
     */
    virtual double asDouble(unsigned int index) const;

    /**
     * Returns single value as double
     * @param index Index of the joint value
     * @return Actual value
     */
    virtual nemo::RealVector asDoubleVector() const;

protected:
    /**
     * Constructor with empty initialization.
     */
    ContainingDoubles();

    /**
     * Constructor
     * Initializes DTO with empty values vector of given dimension.
     */
    ContainingDoubles(unsigned int dimension);

    /**
     * Constructor
     * Initializes DTO with specified double values.
     */
    ContainingDoubles(nemo::RealVector values);

    /**
     * Destructor
     */
    virtual ~ContainingDoubles();

    /**
     * Dimension / number of contained values
     */
    unsigned int _dimension;

    /**
     * Stored double values
     */
    nemo::RealVector _values;
};

}
