#include "CCANode.h"

using namespace ff;
using namespace std;
using namespace boost;

namespace cca {

CCANode::CCANode(string nodename, bool timed, unsigned int samplerate) :
    ff_node(), _nodename(nodename), _timed(timed), _sampleRate(samplerate),
            _taskcounter(0), _svccounter(0), _sentCounter(0), _ticksCounter(0),
            _ticksCounterAll(0), _ready(false), _buffer(), _bufferOld() {
}

CCANode::~CCANode() {
}

void* CCANode::svc(void *task) {
    if (task == EOS) {
        cout << " ++++++++ Node >" << _nodename << "< sending EOS." << endl;
        this->process(EOS);
        return EOS;
    }

    if (!this->_timed) {
        // Node is not timed, we return the processing result right away
        return this->process(task);
    } else {
        // Node is timed, buffer the result
        _buffer = this->process(task);
    }

    return GO_ON;
}

string CCANode::getName() {
    return this->_nodename;
}

void CCANode::tick() {
    //cout << "Node " << this->_nodename << " received tick ... ";
    if (this->_timed) {
        ++this->_ticksCounterAll; // Count received ticks
        ++this->_ticksCounter; // Count received ticks since last processing
        if (this->_ticksCounter >= this->_sampleRate) {
            this->onBeat();
            if (this->_buffer && this->_buffer != NULL) {
                ++this->_sentCounter;
                ff_send_out(this->_buffer);
                this->_bufferOld = this->_buffer; // Always buffering the last command
                this->_buffer = NULL;
            } else {
                // Buffer is empty, so just repeat the last sent item
                if (this->_bufferOld && this->_bufferOld != NULL) {
                    ff_send_out(this->_bufferOld);
                } else {
                    cerr << "Nothing to send." << endl;
                }
            }
            this->_ticksCounter = 0; // Reset ticks counter
        }
    } else {
        cerr << "Node not timed. Ignoring." << endl;
    }
}

void CCANode::onBeat() {
}

void* CCANode::process(void *task) {
    ComplexTask *ct;

    // Timed node receives and processes tasks
    if (DF_DEBUG && DF_VERBOSITY > 1) {
        cout << ">" << _nodename << "< (timed) received Task #" << _taskcounter
                << endl;
    }

    // Check, if just GO_ON without doing anything
    if (task == GO_ON || task == NULL) {
        return GO_ON;
    }

    Task *t = static_cast<Task*> (task);

    if (t->empty()) {
        // Empty task, nothing to do
        if (DF_DEBUG && DF_VERBOSITY > 1) {
            cout << ">" << _nodename << "< (timed) received empty Task #"
                    << _taskcounter << ", returning GO_ON (tc: "
                    << _taskcounter << ")" << endl;
        }

        // FastFlow macro to go on with data-flow without doing anything
        return GO_ON;
    } else {
        // Valid task, process
        ct = static_cast<ComplexTask*> (task);
        if (DF_DEBUG && DF_VERBOSITY > 1) {
            cout << ">" << _nodename << "< (timed) received valid Task #"
                    << _taskcounter << endl;
        }

    }

    if (_timed) {
        // Timed node, just buffering task and sending GO_ON
        _buffer = ct; // Buffer the processed task for next sendout
        ++_taskcounter;
        if (DF_DEBUG && DF_VERBOSITY > 1) {
            cout << ">" << _nodename << "< (timed) buffered Task #"
                    << _taskcounter << endl;
        }
        return GO_ON;
    } else {
        ++_sentCounter;
        if (DF_DEBUG && DF_VERBOSITY > 1) {
            cout << ">" << _nodename << "< sent Task #" << _taskcounter << endl;
        }
        return ct;
    }
}

bool CCANode::isReady() const {
    return this->_ready;
}

bool CCANode::isTimed() const {
    return this->_timed;
}

ostream& operator<<(ostream& os, const CCANode& node) {
    os.precision(3); // Precision when printing double values
    os << node.print();
    return os;
}

}
