#include "MCSP.h"

using namespace ff;

namespace cca {

MCSPNode::MCSPNode(std::string nodename, unsigned int max_input_nodes,
        bool timed, unsigned int samplerate) :
    CCANode(nodename, timed, samplerate), _nInputNodes(0),
            _maxInputNodes(max_input_nodes), _inputNodes(), _prepared(false),
            _outputBuffer(NULL) {
    this->_inputNodes.reserve(max_input_nodes);
}

MCSPNode::~MCSPNode() {
}

void MCSPNode::prepare() {
    // Input buffer
    if (DF_DEBUG && DF_VERBOSITY > 2) {
        std::cout << "MCSPNode>" << _nodename << "<::prepare" << std::endl;
    }
    if (!_prepared) {
        for (unsigned int i = 0; i < _inputNodes.size(); ++i) {
            // Not for the first node, since this will be handled by FastFlow
            // iteself
            if (i != 0) {
                _inputNodes[i]->create_output_buffer(1);
            }
            _inputNodes[i]->svc_init();
            if (DF_DEBUG && DF_VERBOSITY > 2) {
                std::cout << "MCSPNode>" << _nodename
                        << "< prepared output buffers for input node >"
                        << _inputNodes[i]->getName() << "<" << std::endl;
            }
        }
        _prepared = true;
    } else {
        if (DF_DEBUG && DF_VERBOSITY > 2) {
            std::cout << " already prepared" << std::endl;
        }
    }
}

void* MCSPNode::process(void *task) {
    std::cout << "MCSPNode>" << _nodename << "<::process" << std::endl;

    if (task == EOS) {
        std::cout << " ++++ MCSPNode >" << _nodename << "< received EOS."
                << std::endl;
        return EOS;
    }

    void *returntask;
    returntask = this->collectTasks();

    return returntask;
}

void MCSPNode::registerInputNode(CCANode* inputNode) {
    if (_nInputNodes < _maxInputNodes) {
        _inputNodes.push_back(inputNode);
        ++_nInputNodes;
        if (DF_DEBUG && DF_VERBOSITY > 2) {
            std::cout << ">" << _nodename << "< added input node >"
                    << inputNode->getName() << "<" << std::endl;
        }

    } else {
        std::cerr << "MCSPNode>" << _nodename
                << "< can't add any more input nodes." << std::endl;
    }
}

void MCSPNode::registerInputNode(unsigned int position, CCANode* inputNode) {
    if (position < _maxInputNodes) {
        _inputNodes[position] = inputNode;
        // ++_nInputNodes; // TODO:
        if (DF_DEBUG && DF_VERBOSITY > 2) {
            std::cout << "MCSPNode>" << _nodename << "< added input node >"
                    << inputNode->getName() << "< at position #" << position
                    << std::endl;
        }
    } else {
        std::cerr << "MCSPNode>" << _nodename
                << "< can't add input nodes at position #" << position
                << std::endl;
    }
}

void MCSPNode::registerInputNodes(std::vector<CCANode*> &inputNodes) {
    for (unsigned int i = 0; i < inputNodes.size(); ++i) {
        this->registerInputNode(inputNodes[i]);
    }
}

void* MCSPNode::collectTasks() {
    std::cout << "MCSPNode>" << _nodename << "<::collectTasks" << std::endl;
    usleep(50000);

    void *crt;
    void **crtpointer = &crt;

    if (!_prepared) {
        this->prepare();
    }

    for (unsigned int i = 0; i < _inputNodes.size(); ++i) {
        FFBUFFER *outp = _inputNodes[i]->get_out_buffer();

        if (DF_DEBUG && DF_VERBOSITY > 2) {
            if (outp->empty()) {
                std::cout << "MCSPNode>>" << _nodename
                        << "< output buffer of node >"
                        << _inputNodes[i]->getName() << "< empty." << std::endl;

                // FastFlow macro to go on with data-flow without doing anything
                return GO_ON;
            }
        }

        //
        ++_taskcounter;
        _inputNodes[i]->get(crtpointer);

        // Try to cast
        Task *t;
        t = static_cast<Task*> (*crtpointer);
        if (t->empty()) {
            std::cerr << "MCSPNode>" << _nodename
                    << "< collected empty task from node #" << i << "."
                    << std::endl;
        } else {
            ComplexTask *ct;
            ct = static_cast<ComplexTask*> (*crtpointer);
            std::cout << "MCSPNode>>" << _nodename
                    << "< collected task with content >" << ct->getValue(0)
                    << "< from node >" << _inputNodes[i]->getName() << "<"
                    << std::endl;
        }

    }

    return *crtpointer;
}

std::ostream& operator<<(std::ostream& os, const MCSPNode& node) {
    os.precision(3); // Precision when printing double values
    os << std::endl << "MCSPNode >" << node._nodename << "<" << std::endl;
    os << " called " << node._svccounter << " times" << std::endl;
    os << " received " << node._taskcounter << " tasks" << std::endl;
    os << " received from " << node._nInputNodes << " registered inputs"
            << std::endl;
    return os;
}

}
